from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission

from allianceauth.eveonline.models import EveCharacter, EveAllianceInfo, EveCorporationInfo

from charlink.app_imports.utils import AppImport, LoginImport
from esi.models import Token
from app_utils.allianceauth import users_with_permission

from .models import AllianceToken, CorporationToken
from .tasks import update_alliance_contacts, update_corporation_contacts


alliance_scopes = ['esi-alliances.read_contacts.v1']
corporation_scopes = ['esi-corporations.read_contacts.v1']


def _alliance_login(token: Token):
    char = EveCharacter.objects.get(character_id=token.character_id)

    assert char.alliance_id is not None

    try:
        alliance = char.alliance
    except EveAllianceInfo.DoesNotExist:
        alliance = EveAllianceInfo.objects.create_alliance(char.alliance_id)

    assert not AllianceToken.objects.filter(alliance=alliance).exists()

    AllianceToken.objects.create(alliance=alliance, token=token)
    update_alliance_contacts.delay(alliance.alliance_id)


def _corporation_login(token: Token):
    char = EveCharacter.objects.get(character_id=token.character_id)

    try:
        corporation = char.corporation
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(char.corporation_id)

    assert not CorporationToken.objects.filter(corporation=corporation).exists()

    CorporationToken.objects.create(corporation=corporation, token=token)
    update_corporation_contacts.delay(corporation.corporation_id)


def _alliance_users_with_perms():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='aa_contacts',
            codename='manage_alliance_contacts'
        )
    )


def _corporation_users_with_perms():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='aa_contacts',
            codename='manage_corporation_contacts'
        )
    )


app_import = AppImport(
    "aa_contacts",
    [
        LoginImport(
            app_label="aa_contacts",
            unique_id="alliance",
            field_label="Alliance Contacts",
            add_character=_alliance_login,
            scopes=alliance_scopes,
            check_permissions=lambda user: user.has_perm('aa_contacts.manage_alliance_contacts'),
            is_character_added=lambda char: AllianceToken.objects.filter(token__character_id=char.character_id).exists(),
            is_character_added_annotation=Exists(
                AllianceToken.objects.filter(
                    token__character_id=OuterRef('character_id'),
                )
            ),
            get_users_with_perms=_alliance_users_with_perms
        ),
        LoginImport(
            app_label="aa_contacts",
            unique_id="corporation",
            field_label="Corporation Contacts",
            add_character=_corporation_login,
            scopes=corporation_scopes,
            check_permissions=lambda user: user.has_perm('aa_contacts.manage_corporation_contacts'),
            is_character_added=lambda char: CorporationToken.objects.filter(token__character_id=char.character_id).exists(),
            is_character_added_annotation=Exists(
                CorporationToken.objects.filter(
                    token__character_id=OuterRef('character_id'),
                )
            ),
            get_users_with_perms=_corporation_users_with_perms
        )
    ]
)



from django.contrib import admin

from .models import AllianceContact


@admin.register(AllianceContact)
class AllianceContactAdmin(admin.ModelAdmin):
    exclude = ('contact_id', )
    readonly_fields = ('alliance', 'contact_type', 'standing', 'labels', )



from django.db import models
from django.utils import timezone

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter, EveCorporationInfo, EveFactionInfo
from esi.models import Token


class ContactTokenQueryset(models.QuerySet):
    def with_valid_tokens(self):
        valid_tokens = Token.objects.all().require_valid()
        return self.filter(token__in=valid_tokens)


class ContactTokenManager(models.Manager):
    def get_queryset(self):
        return ContactTokenQueryset(self.model, using=self._db)

    def with_valid_tokens(self):
        return self.get_queryset().with_valid_tokens()


class General(models.Model):
    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ('manage_alliance_contacts', 'Can manage alliance contacts'),
            ('manage_corporation_contacts', 'Can manage corporation contacts'),
            ('view_alliance_notes', 'Can view notes on alliance contacts'),
            ('view_corporation_notes', 'Can view notes on corporation contacts'),
        )


class ContactLabel(models.Model):
    label_id = models.BigIntegerField()
    label_name = models.CharField(max_length=255)

    class Meta:
        abstract = True
        default_permissions = ()


class Contact(models.Model):
    contact_id = models.BigIntegerField()

    class ContactTypeOptions(models.TextChoices):
        CHARACTER = 'character'
        CORPORATION = 'corporation'
        ALLIANCE = 'alliance'
        FACTION = 'faction'

    contact_type = models.CharField(max_length=11, choices=ContactTypeOptions.choices)
    standing = models.FloatField()
    notes = models.TextField(blank=True, default='')

    class Meta:
        abstract = True
        default_permissions = ()

    @property
    def image_src(self) -> str:
        if self.contact_type == self.ContactTypeOptions.CHARACTER:
            return EveCharacter.generic_portrait_url(self.contact_id)
        if self.contact_type == self.ContactTypeOptions.CORPORATION:
            return EveCorporationInfo.generic_logo_url(self.contact_id)
        if self.contact_type == self.ContactTypeOptions.ALLIANCE:
            return EveAllianceInfo.generic_logo_url(self.contact_id)
        if self.contact_type == self.ContactTypeOptions.FACTION:
            return EveFactionInfo.generic_logo_url(self.contact_id)
        return ''

    @property
    def contact_name(self) -> str:
        if self.contact_type == self.ContactTypeOptions.CHARACTER:
            try:
                res = EveCharacter.objects.get(character_id=self.contact_id).character_name
            except EveCharacter.DoesNotExist:
                char = EveCharacter.objects.create_character(self.contact_id)
                res = char.character_name
        elif self.contact_type == self.ContactTypeOptions.CORPORATION:
            try:
                res = EveCorporationInfo.objects.get(corporation_id=self.contact_id).corporation_name
            except EveCorporationInfo.DoesNotExist:
                corp = EveCorporationInfo.objects.create_corporation(self.contact_id)
                res = corp.corporation_name
        elif self.contact_type == self.ContactTypeOptions.ALLIANCE:
            try:
                res = EveAllianceInfo.objects.get(alliance_id=self.contact_id).alliance_name
            except EveAllianceInfo.DoesNotExist:
                alliance = EveAllianceInfo.objects.create_alliance(self.contact_id)
                res = alliance.alliance_name
        elif self.contact_type == self.ContactTypeOptions.FACTION:
            try:
                res = EveFactionInfo.objects.get(faction_id=self.contact_id).faction_name
            except EveFactionInfo.DoesNotExist:
                faction = EveFactionInfo.provider.get_faction(self.contact_id)
                EveFactionInfo.objects.create(faction_id=faction.id, faction_name=faction.name)
                res = faction.name
        else:
            res = ''

        return res


class ContactToken(models.Model):
    token = models.ForeignKey(Token, on_delete=models.CASCADE, related_name='+')

    last_update = models.DateTimeField(default=timezone.now)

    objects = ContactTokenManager()

    class Meta:
        abstract = True
        default_permissions = ()


class AllianceContactLabel(ContactLabel):
    alliance = models.ForeignKey(EveAllianceInfo, on_delete=models.RESTRICT, related_name='contact_labels')

    class Meta:
        default_permissions = ()

    def __str__(self):
        return f"{self.alliance} - {self.label_name}"


class AllianceContact(Contact):
    alliance = models.ForeignKey(EveAllianceInfo, on_delete=models.RESTRICT, related_name='contacts')

    labels = models.ManyToManyField(AllianceContactLabel, blank=True, related_name='contacts')

    class Meta:
        default_permissions = ()

    def __str__(self):
        return f"{self.alliance} - {self.contact_name}"


class AllianceToken(ContactToken):
    alliance = models.OneToOneField(EveAllianceInfo, on_delete=models.RESTRICT, related_name='+')

    class Meta:
        default_permissions = ()

    @classmethod
    def visible_for(cls, user):
        if user.is_superuser:
            return cls.objects.all()

        return cls.objects.filter(
            alliance__alliance_id__in=CharacterOwnership.objects
            .filter(user=user)
            .values('character__alliance_id')
        )


class CorporationContactLabel(ContactLabel):
    corporation = models.ForeignKey(EveCorporationInfo, on_delete=models.RESTRICT, related_name='contact_labels')

    class Meta:
        default_permissions = ()

    def __str__(self):
        return f"{self.corporation} - {self.label_name}"


class CorporationContact(Contact):
    corporation = models.ForeignKey(EveCorporationInfo, on_delete=models.RESTRICT, related_name='contacts')

    labels = models.ManyToManyField(CorporationContactLabel, blank=True, related_name='contacts')

    is_watched = models.BooleanField(null=True, blank=True, default=None)

    class Meta:
        default_permissions = ()

    def __str__(self):
        return f"{self.corporation} - {self.contact_name}"


class CorporationToken(ContactToken):
    corporation = models.OneToOneField(EveCorporationInfo, on_delete=models.RESTRICT, related_name='+')

    class Meta:
        default_permissions = ()

    @classmethod
    def visible_for(cls, user):
        if user.is_superuser:
            return cls.objects.all()

        return cls.objects.filter(
            corporation__corporation_id__in=CharacterOwnership.objects
            .filter(user=user)
            .values('character__corporation_id')
        )



# Generated by Django 4.2.8 on 2024-02-13 17:18

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('esi', '0012_fix_token_type_choices'),
        ('eveonline', '0017_alliance_and_corp_names_are_not_unique'),
    ]

    operations = [
        migrations.CreateModel(
            name='General',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'permissions': (('view_contacts', 'Can view alliance contacts'),),
                'managed': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='AllianceToken',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_update', models.DateTimeField(default=django.utils.timezone.now)),
                ('alliance', models.OneToOneField(on_delete=django.db.models.deletion.RESTRICT, related_name='+', to='eveonline.eveallianceinfo')),
                ('token', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='esi.token')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='AllianceContactLabel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label_id', models.BigIntegerField()),
                ('label_name', models.CharField(max_length=255)),
                ('alliance', models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, related_name='contact_labels', to='eveonline.eveallianceinfo')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='AllianceContact',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('contact_id', models.BigIntegerField()),
                ('contact_type', models.CharField(choices=[('character', 'Character'), ('corporation', 'Corporation'), ('alliance', 'Alliance'), ('faction', 'Faction')], max_length=11)),
                ('standing', models.FloatField()),
                ('notes', models.TextField(blank=True, default='')),
                ('alliance', models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, related_name='contacts', to='eveonline.eveallianceinfo')),
                ('labels', models.ManyToManyField(related_name='contacts', to='aa_contacts.alliancecontactlabel')),
            ],
            options={
                'default_permissions': (),
            },
        ),
    ]



from allianceauth import hooks
from allianceauth.services.hooks import UrlHook, MenuItemHook

from . import urls


class AAContactsHook(MenuItemHook):
    def __init__(self):
        super().__init__("Contacts", "fas fa-address-book", "aa_contacts:index", navactive=['aa_contacts:'])


@hooks.register('menu_item_hook')
def register_menu():
    return AAContactsHook()


@hooks.register('url_hook')
def register_urls():
    return UrlHook(urls, 'aa_contacts', 'aa_contacts/')


@hooks.register('charlink')
def register_charlink_hook():
    return 'aa_contacts.charlink_hook.app_import'
